// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

CU.import("resource:///modules/FlockScheduler.jsm");
CU.import("resource:///modules/FlockSvcUtils.jsm");
CU.import("resource:///modules/FlockXPCOMUtils.jsm");
FlockXPCOMUtils.debug = false;
CU.import("resource://gre/modules/JSON.jsm");
CU.import("resource:///modules/FlockXMLUtils.jsm");

const MODULE_NAME = "Digg";

const API_CLASS_NAME = "Flock Digg API";
const API_CLASS_ID = Components.ID("{fd864ab8-d05e-4640-99bf-d4733982c894}");
const API_CONTRACT_ID = "@flock.com/webservice/api/digg;1";
const APP_KEY = "http%3A%2F%2Fflock.com";

const CLASS_NAME = "Flock Digg Service";
const CLASS_SHORT_NAME = "digg";
const CLASS_TITLE = "Digg";
const CLASS_ID = Components.ID("{184dbbf0-ce2d-11dc-95ff-0800200c9a66}");
const CONTRACT_ID = "@flock.com/webservice/digg;1";

const DIGG_NO_AVATAR_ICON = "http://digg.com/img/udl.png";
const FLOCK_NO_AVATAR_ICON = "chrome://flock/skin/people/no_avatar.png";
const DIGG_PROPERTIES = "chrome://flock/locale/services/digg.properties";
const DIGG_URL = "http://digg.com/";
const DOMAIN = "digg.com";

const DIGG_READER_CSS_URL = "chrome://flock/skin/services/digg/streamReader.css";

const FAVICON = "chrome://flock/content/services/digg/favicon.png";
const FLOCK_RDNDS_CONTRACTID = "@flock.com/rich-dnd-service;1";

// Interval to check to do a batch fetch friend.
const FRIEND_INTERVAL_FETCH = 3600;
// The first time, only get photos not older than one week.
const MEDIA_INITIAL_FETCH = 7 * 24 * 3600;

// The delay between two refreshes when the sidebar is closed (in seconds)
const REFRESH_INTERVAL = 1800; // seconds == 30 minutes
// The delay between two refreshes when the sidebar is open (in seconds)
const SHORT_INTERVAL = 300; // seconds == 5 minutes

const SERVICE_ENABLED_PREF = "flock.service.digg.enabled";

// For use with the scheduler
var gTimers = [];

var gApi = null;

var gDuggCurrentUrl = null;

function loadLibraryFromSpec(aSpec) {
  CC["@mozilla.org/moz/jssubscript-loader;1"]
    .getService(CI.mozIJSSubScriptLoader)
    .loadSubScript(aSpec);
}

function createEnum(aEnumArray) {
  return {
    QueryInterface: function createEnum_QueryInterface(aIID) {
      if (aIID.equals(CI.nsISimpleEnumerator)) {
        return this;
      }
      throw CR.NS_ERROR_NO_INTERFACE;
    },
    hasMoreElements: function createEnum_hasMoreElements() {
      return (aEnumArray.length > 0);
    },
    getNext: function createEnum_getNext() {
      return aEnumArray.shift();
    }
  };
}

// This tracking code gets added as a parameter to Digg story links
gTrackingCode = "otc-fl";

/** Adds the Flock tracking code to the specified URL. */
function addTrackingCode(aUrl) {
  return aUrl + ((aUrl.indexOf("?") > -1)
                 ? "&"
                 : "?")
              + gTrackingCode;
}

var flockMediaItemFormatter = {
  canBuildTooltip: true,

  buildTooltip: function fmif_buildTooltip(aMediaItem) {

    default xml namespace =
      "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

    if (aMediaItem.thumbnail.match(/chrome:\/\//)) {
      xml =
        <hbox>
          <vbox>
            <image src={aMediaItem.icon} />
            <spacer flex="1" />
          </vbox>
          <vbox anonid="ttInfoContainer">
            <label anonid="ttTitleTxt">{aMediaItem.title}</label>
            <vbox anonid="ttUserTxt">
              <!-- XXX: localize -->
              <label>Diggs: {aMediaItem.rating_count}</label>
              <label>Comments: {aMediaItem.comment_count}</label>
              <label class="user">{aMediaItem.username}</label>
            </vbox>
          </vbox>
        </hbox>;
    } else {
      xml =
        <vbox>
          <hbox>
            <image src={aMediaItem.largeSizePhoto} style="margin-bottom: 2px;" />
            <spacer flex="1" />
          </hbox>
          <hbox>
            <vbox>
              <image src={aMediaItem.icon} />
              <spacer flex="1" />
            </vbox>
            <vbox anonid="ttInfoContainer">
              <label anonid="ttTitleTxt">{aMediaItem.title}</label>
              <vbox anonid="ttUserTxt">
                <!-- XXX: localize -->
                <label>Diggs: {aMediaItem.rating_count}</label>
                <label>Comments: {aMediaItem.comment_count}</label>
                <label class="user">{aMediaItem.username}</label>
              </vbox>
            </vbox>
          </hbox>
        </vbox>;
      }

    return xml;
  }
};

/*************************************************************************
 * Component: DiggAPI
 *************************************************************************/
function DiggAPI() {
  FlockSvcUtils.getLogger(this);
  this._logger.init("DiggAPI");

  this._WebDetective = CC["@flock.com/web-detective;1"]
                       .getService(CI.flockIWebDetective);
  this._ppUtils = CC["@flock.com/people-utils;1"]
                  .getService(CI.flockIPeopleUtils);
}

DiggAPI.prototype = new FlockXPCOMUtils.genericComponent(
  API_CLASS_NAME,
  API_CLASS_ID,
  API_CONTRACT_ID,
  DiggAPI,
  0,
  [
    CI.flockIDiggAPI,
    CI.flockIWebServiceAPI
  ]
);

DiggAPI.prototype.shortName = "DiggAPI";


/*************************************************************************
 * DiggAPI: flockIDiggAPI Implementation
 *************************************************************************/

/**
 * void getDiggInformation(in AString aURL,
                           in flockIListener aListener,
                           in PRBool aShouldGetTopStories);
 * @see flockIDiggAPI#getDiggInformation
 */
DiggAPI.prototype.getDiggInformation =
function DiggAPI_getDiggInformation(aURL,
                                    aListener,
                                    aShouldGetTopStories)
{
  this._logger.info(".getDiggInformation('" + aURL + "', ...)");
  var inst = this;
  var storyInfo = CC["@mozilla.org/hash-property-bag;1"]
                  .createInstance(CI.nsIWritablePropertyBag2);
/*
  var getDiggInfo_Listener = {
    onSuccess: function getDiggInfo_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getDiggInfo_onSuccess(..., '" + aTopic + "')");
*/
      var wm = CC["@mozilla.org/appshell/window-mediator;1"]
               .getService(CI.nsIWindowMediator);
      var win = wm.getMostRecentWindow("navigator:browser");
/*
      if (aSubject.wrappedJSObject.stories) {
        // There could be several stories related to this URL; We will only
        // deal with the most recent one.
        var story = aSubject.wrappedJSObject.stories[0];
        storyInfo.setPropertyAsAString("storyId", story.id);
        storyInfo.setPropertyAsAString("diggs", story.diggs);
        storyInfo.setPropertyAsAString("submittedDate", story.submit_date);
        storyInfo.setPropertyAsAString("comments", story.comments);
        storyInfo.setPropertyAsAString("user_name", story.user.name)
        var fullname =
          (story.user.fullname) ? story.user.fullname : story.user.name;
        storyInfo.setPropertyAsAString("user_fullname", fullname)
        storyInfo.setPropertyAsAString("user_avatar", story.user.icon);
        var userUrl = inst._WebDetective
                          .getString(CLASS_SHORT_NAME, "profileURL", null)
                          .replace("%accountid%", story.user.name);
        var duggUrl = addTrackingCode(story.href);
        storyInfo.setPropertyAsAString("user_url", userUrl);
        storyInfo.setPropertyAsAString("diggUrl", duggUrl);
        storyInfo.setPropertyAsAString("link", duggUrl);
        gDuggCurrentUrl = duggUrl;
 
        // Detect user who submitted URL
        var pageDoc = win.document.getElementById("content").contentDocument;

        var mediaArr = [];
        var media = {
          name: fullname,
          query: "user:" + story.user.name + "|username:" + fullname,
          label: fullname,
          favicon: FAVICON,
          service: "digg" 
        };
        mediaArr.push(media);

        if (!pageDoc._flock_decorations) {
          pageDoc._flock_decorations = {};
        }

        if (pageDoc._flock_decorations.mediaArr) {
          pageDoc._flock_decorations.mediaArr =
            pageDoc._flock_decorations.mediaArr.concat(mediaArr);
        } else {
          pageDoc._flock_decorations.mediaArr = mediaArr;
        }

        var obs = CC["@mozilla.org/observer-service;1"]
                  .getService(CI.nsIObserverService);
        obs.notifyObservers(pageDoc, "media", "media:update");
      } else {
*/
        // This URL hasn't been dugg
        inst._logger.debug("URL has not been dugg");
        storyInfo.setPropertyAsAString("diggs", "0");
        var encodedUrl = encodeURIComponent(aURL);
        var url = inst._WebDetective
                      .getString(CLASS_SHORT_NAME, "submitUrl", null)
                      .replace("%url%", encodedUrl);

        // Prepopulate title in digg submission
        if (win) {
          url += "&title=" + win.gBrowser.contentTitle;
        }
        // Prepopulate breadcrumb in description body
        var breadcrumb = CC[FLOCK_RDNDS_CONTRACTID]
                         .getService(CI.flockIRichDNDService)
                         .getBreadcrumb("plain");
        if (breadcrumb) {
          url += "&bodytext=" + encodeURIComponent(breadcrumb);
        }

        storyInfo.setPropertyAsAString("diggUrl", url);
        gDuggCurrentUrl = url;
      //}
      if (aShouldGetTopStories) {
        inst._getTopStories(storyInfo, aListener);
      } else {
        aListener.onSuccess(storyInfo, "getDiggInformation");
      }
/*
    },
    onError: function getDiggInfo_onError(aSubject, aTopic, aError) {
      inst._logger.error("getDiggInfo_onError(..., '" + aTopic + ", ...)");
      aListener.onError(aSubject, "getDiggInformation", aError);
    }
  };
  var params = {
    wrappedJSObject: { link: aURL }
  };
  this.call("stories", params, getDiggInfo_Listener);
*/
};

/**
 * void isFriend(in AString aUsername,
 *               in AString aFriendUsername,
 *               in flockIListener aListener);
 * @see flockIDiggAPI#isFriend
 */
DiggAPI.prototype.isFriend =
function DiggAPI_isFriend(aUsername, aFriendUsername, aListener) {
  this._logger.info(".isFriend('" + aUsername + "' , '"
                                  + aFriendUsername + "', ...)");
  var inst = this;
  var isFriend_Listener = {
    onSuccess: function isFriend_onSuccess(aSubject, aTopic) {
      inst._logger.debug("isFriend_onSuccess(..., '" + aTopic + "')");
      var friendStatus = CC["@mozilla.org/hash-property-bag;1"]
                         .createInstance(CI.nsIWritablePropertyBag2);
      try {
        var user = aSubject.wrappedJSObject.users[0];
        if (user && user.mutual) {
          friendStatus.setPropertyAsBool("friendStatus", true);
        } else {
          friendStatus.setPropertyAsBool("friendStatus", false);
        }
      } catch (ex) {
        // A user does not exist
        inst._logger.error("isFriend onError " + ex);
        friendStatus.setPropertyAsBool("friendStatus", false);
      }
      aListener.onSuccess(friendStatus, "isFriend");
    },
    onError: function isFriend_onError(aSubject, aTopic, aError) {
      inst._logger.error("isFriend_onError(..., '" + aTopic + "', ...)");
      aListener.onError(null, "isFriend", aError);
    }
  };
  var method = "user/" + aUsername + "/friend/" + aFriendUsername;
  this.call(method, {}, isFriend_Listener);
};


/**
 * void getDiggInfoByCleanTitle(in AString aTitle, in flockIListener aListener);
 * @see flockIDiggAPI#getDiggInfoByCleanTitle
 */
DiggAPI.prototype.getDiggInfoByCleanTitle =
function DiggAPI_getDiggInfoByCleanTitle(aTitle, aListener) {
  this._logger.info(".getDiggInfoByCleanTitle('" + aTitle + "', ...)");
  var inst = this;
  var getDiggInfoByCleanTitle_Listener = {
    onSuccess: function getDiggInfoByCleanTitle_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getDiggInfoByCleanTitle_onSuccess(..., '"
                                                            + aTopic + "')");
      var storyInfo = CC["@mozilla.org/hash-property-bag;1"]
                      .createInstance(CI.nsIWritablePropertyBag2);
      try {
        var story = aSubject.wrappedJSObject.stories[0];
        storyInfo.setPropertyAsAString("link", addTrackingCode(story.href));
        storyInfo.setPropertyAsAString("diggs", story.diggs);
        storyInfo.setPropertyAsAString("comments", story.comments);
      } catch (ex) {
        // This URL has not yet been dugg
        inst._logger.error("getDiggInfoByCleanTitle onError " + ex);
        storyInfo.setPropertyAsAString("diggs", 0);
      }
      aListener.onSuccess(storyInfo, "getDiggInfoByCleanTitle");
    },
    onError: function getDiggInfoByCleanTitle_onError(aSubject,
                                                      aTopic,
                                                      aError)
    {
      inst._logger.error("getDiggInfoByCleanTitle_onError(..., '"
                                                          + aTopic + ", ...)");
      aListener.onError(null, "getDiggInfoByCleanTitle", aError);
    }
  };
  this.call("story/" + aTitle, {}, getDiggInfoByCleanTitle_Listener);
};

/**
 * void getFriends(in AString aUsername, in flockIListener aListener);
 * @see flockIDiggAPI#getFriends
 */
DiggAPI.prototype.getFriends =
function DiggAPI_getFriends(aUsername, aListener) {
  this._logger.debug(".getFriends('" + aUsername + "', ...)");
  var inst = this;

  // This constant determines the number of friends to retrieve via API.
  // We might as well pull the max possible - 100.
  const API_NUM_FRIENDS = 100;
  var method = "user/" + aUsername + "/friends";
  var result = {};
  var resultCount = 0;

  function processFriend(aFriend) {
    var person = CC["@mozilla.org/hash-property-bag;1"]
                 .createInstance(CI.nsIWritablePropertyBag2);
    person.setPropertyAsAString("name", aFriend.name);
    person.setPropertyAsAString("fullname", aFriend.fullname);
    var avatar = "";
    if (aFriend.icon != DIGG_NO_AVATAR_ICON) {
      avatar = aFriend.icon;
    }
    person.setPropertyAsAString("avatar", avatar);
    result[aFriend.name] = person;
    resultCount++;
    inst._logger.debug(" - friend: " + aFriend.name);
  }

  var getMoreFriends_Listener = {
    onSuccess: function api_getMoreFriends_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getMoreFriends_onSuccess(..., '" + aTopic + "')");
      var totalFriends = aSubject.wrappedJSObject.total;
      var myFriends = aSubject.wrappedJSObject.users;
      for (var k in myFriends) {
        processFriend(myFriends[k]);
      }

      if (resultCount == totalFriends) {
        aListener.onSuccess({ wrappedJSObject: result }, "getMoreFriends");
      } else if (resultCount > totalFriends) {
        inst._logger.error("Found too many friends");
        aListener.onError(null, "getMoreFriends", null);
      }
    },
    onError: function api_getMoreFriends_onError(aSubject, aTopic, aError) {
      inst._logger.error("getMoreFriends_onError(..., '" + aTopic + "', ...");
      aListener.onError(null, "getMoreFriends", aError);
    }
  };

  var getFriends_Listener = {
    onSuccess: function api_getFriends_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getFriends_onSuccess(..., '" + aTopic + "')");
      var totalFriends = aSubject.wrappedJSObject.total;
      var myFriends = aSubject.wrappedJSObject.users;
      for (var i in myFriends) {
        processFriend(myFriends[i]);
      }

      var friendsRemaining = totalFriends - API_NUM_FRIENDS;
      if (friendsRemaining > 0) {
        for (var page = 1; page * API_NUM_FRIENDS < totalFriends; page++) {
          var moreParams = {
            wrappedJSObject: {
              count: API_NUM_FRIENDS,
              offset: page * API_NUM_FRIENDS
            }
          };
          inst.call(method, moreParams, getMoreFriends_Listener);
        }
      } else {
        aListener.onSuccess({ wrappedJSObject: result }, "getFriends");
      }
    },
    onError: function api_getFriends_onError(aSubject, aTopic, aError) {
      inst._logger.error("getFriends_onError(..., '" + aTopic + "', ...)");
      aListener.onError(null, "getFriends", aError);
    }
  };

  var params = {
    wrappedJSObject: { count: API_NUM_FRIENDS, offset: 0 }
  };
  this.call(method, params, getFriends_Listener);
};

DiggAPI.prototype._updateRecentUrlForUser =
function DiggAPI__updateRecentUrlForUser(aStoryId, aFriend, aListener) {
  this._logger.debug("._updateRecentUrlForUser('" + aStoryId + "', '"
                                                  + aFriend.accountId + "', ...)");
  var inst = this;
  var updateRecentUrlForUser_Listener = {
    onSuccess: function updateRecentUrlForUser_onSuccess(aSubject, aTopic) {
      inst._logger.debug("updateRecentUrlForUser_onSuccess(..., '"
                         + aTopic + "')");
      // There could be several stories related to this URL; we will only
      // deal with the most recent one.
      var story = aSubject.wrappedJSObject.stories[0];
      if (story) {
        aFriend.statusMessage = story.title;
        aFriend.statusMessageUrl = addTrackingCode(story.href);
        if (story.user.name == aFriend.accountId) {
          aFriend.lastUpdateType = "submitted";
        }
        if (aFriend.lastUpdateType == "dugg") {
          aFriend.statusMessage += " (" + story.diggs + ")";
        }
      }
      aListener.onSuccess(aFriend, "_updateRecentUrlForUser");
    },
    onError: function updateRecentUrlForUser_onError(aSubject, aTopic, aError) {
      inst._logger.error("updateRecentUrlForUser_onError('" + aTopic + "')");
      aListener.onError(aFriend, "", aError);
    }
  };
  this.call("stories/" + aStoryId, null, updateRecentUrlForUser_Listener);
}

/**
 * void getBatchFriendActivity(in nsISupports aFriendsList,
 *                             in flockIListener aListener);
 * @see flockIDiggAPI#getBatchFriendActivity
 */
DiggAPI.prototype.getBatchFriendActivity =
function DiggAPI_getBatchFriendActivity(aAccount, aFriendsList, aListener) {
  var inst = this;
  var storyIds = [];
  var friendsName = "";
  var friendDiggGetter = {
    listener: {
      onSuccess: function gbfa_fdg_listener_onSuccess(aSubject, aTopic) {
        inst._logger.info("gbfa friendDiggGetter.listener.onSuccess ");
        var diggsLastRetrievedDate = aSubject.wrappedJSObject.timestamp;
        aAccount.setCustomParam("diggsLastRetrievedDate", diggsLastRetrievedDate);
        var diggs = aSubject.wrappedJSObject.diggs;
        for (var d in diggs) {
          var digg = diggs[d];
          var friend = aFriendsList.getProperty(digg.user);
          if (digg.date > friend.lastUpdate) {
            friend.lastUpdate = digg.date;
            friend.lastUpdateType = "dugg";
            friend.unseenMedia++;
            storyIds[digg.story] = friend;
            aFriendsList.setProperty(digg.user, friend);
          }
        }
        friendDiggGetter.updateAllUsersStories();
      },
      onError: function gbfa_fdg_listener_onError(aSubject,
                                                   aTopic,
                                                   aError)
      {
        inst._logger.error("gbfa friendDiggGetter.listener.onError(..., '"
                           + aTopic + "')");
        friendDiggGetter._updateAllUsersStories();
      }
    },
    getDiggs: function gbfa_fdg_getDiggs() {
      var method = "users/" + friendsName + "/diggs";
      // If we have retrieved diggs before then don't bother asking for items
      // earlier than this date. We do this by setting the min_date parameter in
      // our request for diggs. If we have not asked for diggs before then we
      // don't pass a min_date which gets all diggs in the past hour.
      var diggsLastRetrievedDate = aAccount.getCustomParam("diggsLastRetrievedDate");
      var params = null;
      if (diggsLastRetrievedDate) {
        params = {
          wrappedJSObject: {
            min_date: diggsLastRetrievedDate
          }
        };
      }

      inst.call(method, params, this.listener);
    },
    updateAllUsersStories: function gbfa_fdg_updateAllUsersStories() {
      for (id in storyIds) {
        var friend = storyIds[id];
        inst._updateUserStory(id, friend, aListener);
      }
    }
  };
  // Get friends' comments.
  var getFriendComment_Listener = {
    onSuccess: function getFriendComment_onSuccess(aSubject, aTopic) {
      inst._logger.info("getFriendComment_onSuccess ");
      try {
        var commentsLastRetrievedDate = aSubject.wrappedJSObject.timestamp;
        aAccount.setCustomParam("commentsLastRetrievedDate", commentsLastRetrievedDate);
        var comments = aSubject.wrappedJSObject.comments;
        var total = parseInt(aSubject.wrappedJSObject.total);
        if (comments && total > 0) {
          for (var c in comments) {
            var comment = comments[c];
            var friend = aFriendsList.getProperty(comment.user);
            if (comment.date > friend.lastUpdate) {
              friend.lastUpdate = comment.date;
              friend.lastUpdateType = "commented";
              storyIds[comment.story] = friend;
              var comments = aAccount.getCustomParam("diggComments");
              aAccount.setCustomParam("diggComments", comments + 1);
              aFriendsList.setProperty(comment.user, friend);
              inst._ppUtils.lightPeopleIcon();
            }
          }
        }
        friendDiggGetter.getDiggs();
      } catch (ex) {
        inst._logger.error("getFriendComment_Listener.onSuccess " + ex);
        friendDiggGetter.getDiggs();
      }
    },
    onError: function getFriendComment_onError(aSubject, aTopic, aError) {
      inst._logger.error("getBatchFriendActivity "
                         + "getFriendComment_onError('" + aTopic + "')");
      friendDiggGetter.getDiggs();
    }
  };
  var friendArray = [];
  var enumerate = aFriendsList.enumerator;
  while (enumerate.hasMoreElements()) {
    var el = enumerate.getNext();
    el.QueryInterface(CI.nsIProperty);
    friendArray.push(el.name)
  }
  friendsName = friendArray.join(",");
  var method = "users/" + friendsName + "/comments";
  var commentsLastRetrievedDate = aAccount.getCustomParam("commentsLastRetrievedDate");
  // If we have retrieved comments before then don't bother asking for items
  // earlier than this date. We do this by setting the min_date parameter in
  // our request for comments. If we have not asked for comments before then we
  // don't pass a min_date which gets all comments in the past hour.
  var params = null;
  if (commentsLastRetrievedDate) {
    params = {
      wrappedJSObject: {
        min_date: commentsLastRetrievedDate
      }
    };
  }
  this.call(method, params, getFriendComment_Listener);
}

DiggAPI.prototype._updateUserStory =
function DiggAPI__updateUserStory(aStoryId, aFriend, aListener) {
  this._logger.info("_updateUserStory('" + aStoryId + "')")
  if (!aStoryId) {
    aListener.onSuccess(aFriend, "getFriendComment");
  } else {
    this._updateRecentUrlForUser(aStoryId, aFriend, aListener);
  }
}

/**
 * void getFriendActivity(in flockIPerson aFriend, in flockIListener aListener);
 * @see flockIDiggAPI#getFriendActivity
 */
DiggAPI.prototype.getFriendActivity =
function DiggAPI_getFriendActivity(aAccount, aFriend, aListener) {
  var inst = this;
  var storyId = null;
  aFriend.QueryInterface(CI.flockIPerson);
  // Get friends' diggs.
  var friendDiggGetter = {
    listener: {
      onSuccess: function gfa_fdg_listener_onSuccess(aSubject, aTopic) {
        inst._logger.info("gfa_fdg_listener_onSuccess");
        var diggs = aSubject.wrappedJSObject.diggs;
        if (diggs) {
          var story = diggs[0];
          if (story.date) {
            aFriend.lastUpdate = story.date;
            aFriend.unseenMedia++;
            aFriend.lastUpdateType = "dugg";
            storyId = story.story;
          }
        }
        inst._updateUserStory(storyId, aFriend, aListener);
      },
      onError: function gfa_fdg_listener_onError(aSubject, aTopic, aError) {
        inst._logger.error("gfa_fdg_listener_onError('" + aTopic + "')");
        inst._updateUserStory(storyId, aFriend, aListener);
      }
    },
    getDiggs: function getDiggs() {
      var method = "users/" + aFriend.accountId + "/diggs"
      var params = {
        wrappedJSObject: { min_date: aFriend.lastUpdate,
                           count: 1 }
      };
      inst.call(method, params, this.listener);
    } 
  };
  var getFriendComment_Listener = {
    onSuccess: function getFriendComment_onSuccess(aSubject, aTopic) {
      var comments = aSubject.wrappedJSObject.comments;
      var totalComments = aSubject.wrappedJSObject.total;
      if (comments && totalComments) {
        totalComments = parseInt(totalComments);
        var story = comments[0];
        if (story.date && story.date > aFriend.lastUpdate) {
          aFriend.lastUpdate = story.date;
          aFriend.lastUpdateType = "commented";
          storyId = story.story;
          var comments = aAccount.getCustomParam("diggComments");
          aAccount.setCustomParam("diggComments", comments + totalComments);
          inst._ppUtils.lightPeopleIcon();
        }
      }
      friendDiggGetter.getDiggs();
    },
    onError: function getFriendComment_onError(aSubject, aTopic, aError) {
      inst._logger.error("getFriendComment_onError('" + aTopic + "')");
      friendDiggGetter.getDiggs();
    }
  };
  var method = "users/" + aFriend.accountId + "/comments";
  var params = {
    wrappedJSObject: { min_date: aFriend.lastUpdate }
  };
  this.call(method, params, getFriendComment_Listener);
}

/**
 * void getFriendComment(in flockIPerson aFriend, in flockIListener aListener);
 * @see flockIDiggAPI#getFriendComment
 */
DiggAPI.prototype.getFriendComment =
function DiggAPI_getFriendComment(aFriend, aListener) {
  var inst = this;
  aFriend.QueryInterface(CI.flockIPerson);
  var getFriendComment_Listener = {
    onSuccess: function getFriendComment_onSuccess(aSubject, aTopic) {
      var story = aSubject.wrappedJSObject.comments[0];
      if (story.date) {
        if (story.date > aFriend.lastUpdate) {
          aFriend.lastUpdate = story.date;
          aFriend.lastUpdateType = "commented"
        }
      }
      aListener.onSuccess(aFriend, "getFriendComment");
    },
    onError: function getFriendComment_onError(aSubject, aTopic, aError) {
      inst._logger.error("getFriendComment_onError('" + aTopic + "')");
      aListener.onError(aFriend, "", aError);
    }
  };
  var method = "users/" + aFriend.accountId + "/comments"
    var params = {
    wrappedJSObject: { min_date: aFriend.lastUpdate,
                       count: 1 }
  };
  this.call(method, params, getFriendComment_Listener);
}

/**
 * void getUser(in AString aUsername, in flockIListener aListener);
 * @see flockIDiggAPI#getUser
 */
DiggAPI.prototype.getUser =
function DiggAPI_getUser(aUsername, aListener) {
  this._logger.debug(".getUser('" + aUsername + "', ...)");
  // This listener handles getting the account owner's information.
  var inst = this;
  var getUser_Listener = {
    onSuccess: function api_getUser_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getUser_onSuccess(..., '" + aTopic + "')");
      var userInfo = CC["@mozilla.org/hash-property-bag;1"]
                     .createInstance(CI.nsIWritablePropertyBag2);
      var user = aSubject.wrappedJSObject.users[0];
      userInfo.setPropertyAsAString("name",
                                    (user.fullname) ? user.fullname
                                                    : aUsername);
      userInfo.setPropertyAsAString("screenName",
                                    (user.name) ? user.name : aUsername);
      userInfo.setPropertyAsAString("avatar", (user.icon) ? user.icon : "");
      var profileURL = inst._WebDetective
                           .getString(CLASS_SHORT_NAME, "profileURL", null)
                           .replace("%accountid%", aUsername);
      userInfo.setPropertyAsAString("URL", profileURL);
      aListener.onSuccess(userInfo, "getUser");
    },
    onError: function api_getUser_onError(aSubject, aTopic, aError) {
      inst._logger.error("getUser_onError(..., '" + aTopic + "', ...)");
      aListener.onError(null, "getUser", aError);
    }
  };
  this.call("user/" + aUsername, null, getUser_Listener);
};

DiggAPI.prototype._getTopStories =
function DiggAPI__getTopStories(aDataObject, aListener) {
  var inst = this;
  var getTopStories_Listener = {
    onSuccess: function api_getTopStories_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getTopStories_onSuccess(..., '" + aTopic + "')");
      var topStories = [];
      try {
        // There could be several stories related to this URL; We will only
        // deal with the most recent one.
        var stories = aSubject.wrappedJSObject.stories;
        for (var s in stories) {
          var story = stories[s];
          topStories.push({
            id: story.id,
            title: story.title,
            origLink: story.href, // Digg url w/o tracking code
            link: addTrackingCode(story.href), // Digg url with tracking code
            url: story.link, // Actual url of Digg story
            diggs: story.diggs
          });
        }
        var result = JSON.toString(topStories);
        aDataObject.setPropertyAsAString("topStories", result);
      } catch (e) {
        inst._logger.debug("getTopStories_onError " + e);
        aListener.onError(null, "getTopStories", e);
        aDataObject.setPropertyAsAString("topStories", "null");
      }
      aListener.onSuccess(aDataObject, "getTopStories");
    },
    onError: function api_getTopStories_onError(aSubject, aTopic, aError) {
      inst._logger.error("getTopStories_onError(..., '" + aTopic + "', ...)");
      aListener.onError(null, "getTopStories", aError);
    }
  };
  var params = {
    wrappedJSObject: { count: 5 }
  };
  this.call("stories/top", params, getTopStories_Listener);
}

/**
 * void getTopStories(in flockIListener aListener);
 * @see flockIDiggAPI#getTopStories
 */
DiggAPI.prototype.getTopStories =
function DiggAPI_getTopStories(aListener) {
  var result = CC["@mozilla.org/hash-property-bag;1"]
               .createInstance(CI.nsIWritablePropertyBag2);
  this._getTopStories(result, aListener);
};

/*************************************************************************
 * DiggAPI: flockIWebServiceAPI Implementation
 *************************************************************************/

/**
 * void call(in AString aMethod,
 *           in nsISupports aParams,
 *           in flockIListener aListener);
 * @see flockIWebServiceAPI#call
 */
DiggAPI.prototype.call =
function DiggAPI_call(aMethod, aParams, aListener) {
  this._logger.debug(".call('" + aMethod + "', ...)");
  var req = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
            .createInstance(CI.nsIXMLHttpRequest)
            .QueryInterface(CI.nsIJSXMLHttpRequest);
  var inst = this;
  req.onreadystatechange = function call_onReadyStateChange() {
    try {
      if (req.readyState == 4) {
        inst._logger.debug(".call('" + aMethod + "', ...) responseText:\n"
                           + req.responseText);
        if (req.status == 200) {
          var s = new CU.Sandbox("about:blank");
          var result = CU.evalInSandbox("(" + req.responseText + ")", s);
          if (result) {
            aListener.onSuccess({ wrappedJSObject: result }, "call");
          } else {
            inst._logger.error(".call() ERROR: no result");
            var xml_error = CI.flockIError.INVALID_RESPONSE_XML;
            aListener.onError(null, "call",
                              inst.getError(xml_error));
          }
        } else {
          aListener.onError(null, "call", null);
        }
      }
    } catch (e) {
      inst._logger.error(".call() ERROR: from server " + e);
      aListener.onError(null, "call", null);
    }
  };
  // TODO: web dic
  var callURL = "http://digg.com/tools/services/?endPoint=/" + aMethod
              + "&appkey=" + APP_KEY
              + "&type=json";
  if (aParams) {
    var params = aParams.wrappedJSObject;
    for (var p in params) {
      callURL += "&" + encodeURIComponent(p)
              + "=" + encodeURIComponent(params[p]);
    }
  }
  this._logger.debug("calling URL: " + callURL);
  req.open("GET", callURL, true);
  req.send(null);
};

/**
 * flockIError getError(in AString aErrorCode);
 * @see flockIWebServiceAPI#getError
 */
DiggAPI.prototype.getError =
function DiggAPI_getError(aErrorCode) {
  var error = CC[FLOCK_ERROR_CONTRACTID].createInstance(CI.flockIError);
  error.errorCode = aErrorCode;
  switch (aErrorCode) {
    case CI.flockIError.INVALID_RESPONSE_XML:
      error.errorString = "responseText did not produce an object";
  }
  return error;
};

/**
 * flockIError getHttpError(in AString aHttpErrorCode);
 * @see flockIWebServiceAPI#getHttpError
 */
DiggAPI.prototype.getHttpError =
function DiggAPI_getHttpError(aHttpErrorCode) {
  return null;
};


/*************************************************************************
 * Component: DiggService
 *************************************************************************/
function DiggService() {
  FlockSvcUtils.getLogger(this);

  loadLibraryFromSpec("chrome://browser/content/flock/photo/photoAPI.js");

  // Determine whether this service has been disabled, and unregister if so.
  var prefService = CC["@mozilla.org/preferences-service;1"]
                    .getService(CI.nsIPrefBranch);
  if (prefService.getPrefType(SERVICE_ENABLED_PREF) &&
     !prefService.getBoolPref(SERVICE_ENABLED_PREF))
  {
    this._logger.debug(SERVICE_ENABLED_PREF + " is FALSE! Not initializing.");
    this.deleteCategories();
    return;
  }

  this._obs = CC["@mozilla.org/observer-service;1"]
              .getService(CI.nsIObserverService);
  this._obs.addObserver(this, "xpcom-shutdown", false);

  this._accountClassCtor = DiggAccount;

  FlockSvcUtils.getCoop(this);
  FlockSvcUtils.getACUtils(this);

  this._baseUrn = "urn:digg";
  this._serviceUrn = this._baseUrn + ":service";

  this._c_svc = this._coop.get(this._serviceUrn);
  if (!this._c_svc) {
    this._c_svc = new this._coop.Service(this._serviceUrn, {
      name: CLASS_SHORT_NAME,
      desc: CLASS_TITLE
    });
  }
  this._c_svc.serviceId = CONTRACT_ID;
  this._c_svc.logoutOption = true;

  
  this._c_svc.domains = FlockSvcUtils.getWD(this).getString(CLASS_SHORT_NAME,
                                                            "domains",
                                                            DOMAIN);
  gTrackingCode = this._WebDetective.getString(CLASS_SHORT_NAME,
                                               "trackingCode",
                                               gTrackingCode);
  this._ppUtils = CC["@flock.com/people-utils;1"]
                  .getService(CI.flockIPeopleUtils);

  // Initialize member that specifies path for localized string bundle
  this._stringBundlePath = DIGG_PROPERTIES;

  // Initialize API
  gApi = new DiggAPI();

  // Update auth states
  try {
    if (this._WebDetective.detectCookies(CLASS_SHORT_NAME,
                                         "loggedoutcookies", null))
    {
      this._acUtils.markAllAccountsAsLoggedOut(CONTRACT_ID);
    }
  } catch (ex) {
    this._logger.error("ERROR updating auth states for Digg: " + ex);
  }

  var ws = FlockSvcUtils.flockIWebService;
  ws.addDefaultMethod(this, "getStringBundle");
  this._loginURL = "http://digg.com/login";
  ws.addDefaultMethod(this, "loginURL");
  ws.addDefaultMethod(this, "isHidden");

  var lws = FlockSvcUtils.flockIManageableWebService;
  lws.addDefaultMethod(this, "docRepresentsSuccessfulLogin");
  ws.addDefaultMethod(this, "getAccount");
  lws.addDefaultMethod(this, "getAccountIDFromDocument");
  ws.addDefaultMethod(this, "getAccounts");
  ws.addDefaultMethod(this, "getAuthenticatedAccount");
  lws.addDefaultMethod(this, "getCredentialsFromForm");
  lws.addDefaultMethod(this, "getSessionValue");
  lws.addDefaultMethod(this, "ownsDocument");
  lws.addDefaultMethod(this, "ownsLoginForm");
  ws.addDefaultMethod(this, "removeAccount");

  var mws = FlockSvcUtils.flockIMediaWebService;
  mws.addDefaultMethod(this, "getChannel");
  mws.addDefaultMethod(this, "enumerateChannels");
  mws.addDefaultMethod(this, "getMediaItemFormatter");
  mws.addDefaultMethod(this, "getIconForQuery");

  var sws = FlockSvcUtils.flockISocialWebService;
  sws.addDefaultMethod(this, "markAllMediaSeen");

  // Default media channels
  var bundle = this.getStringBundle();
  this._channels = {
    "all": {
      title: bundle.GetStringFromName("flock.digg.title.all"),
      supportsSearch: false,
      subchannels: {
        "special:all": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:all-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:all-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:all-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:all-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "technology": {
      title: bundle.GetStringFromName("flock.digg.title.technology"),
      supportsSearch: false,
      subchannels: {
        "special:technology": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:technology-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:technology-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:technology-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:technology-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "world_business": {
      title: bundle.GetStringFromName("flock.digg.title.world_business"),
      supportsSearch: false,
      subchannels: {
        "special:world_business": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:world_business-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:world_business-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:world_business-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:world_business-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "science": {
      title: bundle.GetStringFromName("flock.digg.title.science"),
      supportsSearch: false,
      subchannels: {
        "special:science": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:science-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:science-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:science-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:science-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "gaming": {
      title: bundle.GetStringFromName("flock.digg.title.gaming"),
      supportsSearch: false,
      subchannels: {
        "special:gaming": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:gaming-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:gaming-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:gaming-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:gaming-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "lifestyle": {
      title: bundle.GetStringFromName("flock.digg.title.lifestyle"),
      supportsSearch: false,
      subchannels: {
        "special:lifestyle": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:lifestyle-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:lifestyle-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:lifestyle-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:lifestyle-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "entertainment": {
      title: bundle.GetStringFromName("flock.digg.title.entertainment"),
      supportsSearch: false,
      subchannels: {
        "special:entertainment": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:entertainment-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:entertainment-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:entertainment-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:entertainment-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "sports": {
      title: bundle.GetStringFromName("flock.digg.title.sports"),
      supportsSearch: false,
      subchannels: {
        "special:sports": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:sports-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:sports-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:sports-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:sports-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "offbeat": {
      title: bundle.GetStringFromName("flock.digg.title.offbeat"),
      supportsSearch: false,
      subchannels: {
        "special:offbeat": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:offbeat-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:offbeat-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:offbeat-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:offbeat-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "news": {
      title: bundle.GetStringFromName("flock.digg.title.news"),
      supportsSearch: false,
      subchannels: {
        "special:news": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:news-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:news-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:news-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:news-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "videos": {
      title: bundle.GetStringFromName("flock.digg.title.videos"),
      supportsSearch: false,
      subchannels: {
        "special:videos": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:videos-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:videos-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:videos-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:videos-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    },
    "images": {
      title: bundle.GetStringFromName("flock.digg.title.images"),
      supportsSearch: false,
      subchannels: {
        "special:images": {
          title: bundle.GetStringFromName("flock.digg.title.recent"),
          supportsSearch: false
        },
        "special:images-24h": {
          title: bundle.GetStringFromName("flock.digg.title.24h"),
          supportsSearch: false
        },
        "special:images-7d": {
          title: bundle.GetStringFromName("flock.digg.title.7d"),
          supportsSearch: false
        },
        "special:images-30d": {
          title: bundle.GetStringFromName("flock.digg.title.30d"),
          supportsSearch: false
        },
        "special:images-365d": {
          title: bundle.GetStringFromName("flock.digg.title.365d"),
          supportsSearch: false
        }
      }
    }
  };
}

DiggService.prototype = new FlockXPCOMUtils.genericComponent(
  CLASS_NAME,
  CLASS_ID,
  CONTRACT_ID,
  DiggService,
  CI.nsIClassInfo.SINGLETON,
  [
    CI.flockIDiggService,
    CI.flockIManageableWebService,
    CI.flockIMediaWebService,
    CI.flockIPollingService,
    CI.flockISocialWebService,
    CI.flockIWebService,
    CI.flockIReaderExtension,
    CI.nsIObserver
  ]
);

DiggService.prototype._xpcom_categories = [
  { category: "wsm-startup" },
  { category: "flockMediaProvider", entry: CLASS_SHORT_NAME },
  { category: "flockWebService", entry: CLASS_SHORT_NAME },
  { category: "flockReaderExtension", entry: CONTRACT_ID }
];


/**************************************************************************
 * DiggService: nsIObserver Implementation
 **************************************************************************/

/**
 * void observe(in nsISupports subject, in char* topic, in PRUnichar* data);
 * @see nsIObserver#observe
 */
DiggService.prototype.observe =
function DiggService_observe(aSubject, aTopic, aState) {
  this._logger.info(".observe(..., '" + aTopic + "', '" + aState + "')");
  switch (aTopic) {
    case "xpcom-shutdown":
      this._obs.removeObserver(this, "xpcom-shutdown");
  }
};

DiggService.prototype.isFriend =
function DiggService_isFriend(aUsername, aFriendUsername, aListener) {
  this._logger.info(".isFriend('" + aUsername + "' , '"
                                  + aFriendUsername + "', ...)");
  gApi.isFriend(aUsername, aFriendUsername, aListener);
};

/**************************************************************************
 * DiggService: flockIWebService Implementation
 **************************************************************************/

// readonly attribute AString contractId;
// ALMOST duplicated by nsIClassInfo::contractID
// with different case: contractId != contractID
DiggService.prototype.contractId = CONTRACT_ID;

// readonly attribute AString icon;
DiggService.prototype.icon = FAVICON;

// readonly attribute AString shortName;
DiggService.prototype.shortName = CLASS_SHORT_NAME;

// readonly attribute AString title;
DiggService.prototype.title = CLASS_TITLE;

// readonly attribute AString url;
DiggService.prototype.url = DIGG_URL;

// readonly attribute AString urn;
DiggService.prototype.urn = "urn:" + CLASS_SHORT_NAME + ":service";

// DEFAULT: nsIStringBundle getStringBundle();
// DEFAULT: boolean isHidden();


/**************************************************************************
 * DiggService: flockILoginWebService Implementation
 **************************************************************************/

// readonly attribute AString domains;
// gets set in _init()

// readonly attribute AString loginURL;
// gets set in _init()

// readonly attribute boolean needPassword;
DiggService.prototype.needPassword = true;

/**
 * flockIWebServiceAccount addAccount(in AString aAccountID,
 *                                    in boolean aIsTransient,
 *                                    in flockIListener aListener);
 * @see flockILoginWebService#addAccount
 */
DiggService.prototype.addAccount =
function DiggService_addAccount(aAccountId, aIsTransient, aListener) {
  this._logger.info(".addAccount('" + aAccountId + "', "
                                    + aIsTransient + ", ...)");

  var url = this._WebDetective
                .getString(CLASS_SHORT_NAME, "profileURL", null)
                .replace("%accountid%", aAccountId);
  var accountUrn = this._acUtils.createAccount(this,
                                               aAccountId,
                                               aAccountId,
                                               url,
                                               aIsTransient);
  // Instantiate account component
  var account = this.getAccount(accountUrn);
  account.setParam("isPollable", true);
  account.setParam("refreshInterval", REFRESH_INTERVAL);
  var params = CC["@mozilla.org/hash-property-bag;1"]
               .createInstance(CI.nsIWritablePropertyBag2);
  params.setPropertyAsInt32("lastFriendRefresh", 0);
  params.setPropertyAsInt32("diggComments", 0);
  this._acUtils.addCustomParamsToAccount(params, accountUrn);
  if (aListener) {
    aListener.onSuccess(account, "addAccount");
  }
  return account;
};

// DEFAULT: boolean docRepresentsSuccessfulLogin(in nsIDOMHTMLDocument aDocument);
// DEFAULT: flockIWebServiceAccount getAccount(in AString aAccountURN);
// DEFAULT: AString getAccountIDFromDocument(in nsIDOMHTMLDocument aDocument);
// DEFAULT: nsISimpleEnumerator getAccounts();
// DEFAULT: flockIWebServiceAccount getAuthenticatedAccount();
// DEFAULT: nsILoginInfo getCredentialsFromForm(in nsIDOMHTMLFormElement aForm);
// DEFAULT: AString getSessionValue();

/**
 * void logout();
 * @see flockILoginWebService#logout
 */
DiggService.prototype.logout =
function DiggService_logout() {
  this._logger.debug(".logout()");
  this._acUtils.removeCookies(this._WebDetective
                                  .getSessionCookies(CLASS_SHORT_NAME));
  this._acUtils.markAllAccountsAsLoggedOut(CONTRACT_ID);
  if (gTimers.length > 0) {
    FlockScheduler.cancel(gTimers, 0);
  }
};

// DEFAULT: boolean ownsDocument(in nsIDOMHTMLDocument aDocument);
// DEFAULT: boolean ownsLoginForm(in nsIDOMHTMLFormElement aForm);
// DEFAULT: void removeAccount(in AString aAccountURN);

/**
 * void updateAccountStatusFromDocument(in nsIDOMHTMLDocument aDocument,
 *                                      in AString aAcctURN,
 *                                      in flockIListener aAuthListener);
 * @see flockILoginWebService#updateAccountStatusFromDocument
 */
DiggService.prototype.updateAccountStatusFromDocument =
function DiggService_updateAccountStatusFromDocument(aDocument,
                                                     aAcctURN,
                                                     aAuthListener)
{
  this._logger.info(".updateAccountStatusFromDocument('" + aDocument.URL + "', '"
                                                         + aAcctURN + "', ...)");
  if (aAcctURN) {
    if (!this._acUtils.ensureOnlyAuthenticatedAccount(aAcctURN)) {
      // This account has JUST been logged in
      var account = this.getAccount(aAcctURN);
      account.login(aAuthListener);
    }
  } else if (this._WebDetective
                 .detect(CLASS_SHORT_NAME, "loggedout", aDocument, null))
  {
    this._acUtils.markAllAccountsAsLoggedOut(CONTRACT_ID);
  }
};

DiggService.prototype.getTopStories =
function DiggService_getTopStories(aListener) {
  gApi.getTopStories(aListener);
};

/**************************************************************************
 * DiggService: flockISocialWebService Implementation
 **************************************************************************/

// readonly attribute long maxStatusLength;
DiggService.prototype.maxStatusLength = 0;

// DEFAULT: void markAllMediaSeen(in AString aPersonUrn);


/**************************************************************************
 * DiggService: flockIPageRankingWebService Implementation
 **************************************************************************/

DiggService.prototype.loadDiggInfo =
function DiggService_loadDiggInfo(aURL, aListener, aShouldGetTopStories) {
  this._logger.info(".loadDiggInfo('" + aURL + "', ...)");
  gApi.getDiggInformation(aURL, aListener, aShouldGetTopStories);
};

DiggService.prototype.getDiggInfoByCleanTitle =
function DiggService_getDiggInfoByCleanTitle(aTitle, aListener) {
  this._logger.info(".getDiggInfoByCleanTitle('" + aTitle + "', ...)");
  gApi.getDiggInfoByCleanTitle(aTitle, aListener);
};

/**************************************************************************
 * DiggService: flockIMediaWebService Implementation
 **************************************************************************/

// readonly attribute boolean flockIMediaWebService::supportsUsers
DiggService.prototype.supportsUsers = true;

/**
 * void decorateForMedia(in nsIDOMHTMLDocument aDocument);
 * @see flockIMediaWebService#decorateForMedia
 */
DiggService.prototype.decorateForMedia =
function DiggService_decorateForMedia(aDocument) {
  this._logger.debug(".decorateForMedia()");
  var results = CC["@mozilla.org/hash-property-bag;1"]
                .createInstance(CI.nsIWritablePropertyBag2);
  var inst = this; 
  if (this._WebDetective.detect("digg", "person", aDocument, results)) {
    var userId = results.getPropertyAsAString("userid");
    var decorateListener = {
      onSuccess: function decorateListener_onSuccess(aSubject, aTopic) {
        var person = aSubject.QueryInterface(CI.nsIPropertyBag);
        var username;
        try {
          username = person.getProperty("name");
        } catch (ex) {
          username = userId; 
        }

        var mediaArr = [];
        var media = {
          name: username,
          query: "user:" + userId + "|username:" + username,
          label: username,
          favicon: inst.icon,
          service: inst.shortName
        }
        mediaArr.push(media);
        if (!aDocument._flock_decorations) {
          aDocument._flock_decorations = {};
        }

        if (aDocument._flock_decorations.mediaArr) {
          aDocument._flock_decorations.mediaArr =
            aDocument._flock_decorations.mediaArr.concat(mediaArr);
        } else {
          aDocument._flock_decorations.mediaArr = mediaArr;
        }
        inst._obs.notifyObservers(aDocument, "media", "media:update");
      },
      onError: function decorateListener_onError(aSubject, aTopic, aError) {
        inst._logger.error("getUser_onError(..., '" + aTopic + "', ...)");
        aListener.onError(null, "getUser", aError);
      }
    };
    gApi.getUser(userId, decorateListener);
  }
};

/**
 * void findByUsername(in flockIListener aListener, in AString aUsername);
 * @see flockIMediaWebService#findByUsername
 */
DiggService.prototype.findByUsername =
function DiggService_findByUsername(aListener, aUsername) {
  this._logger.debug(".findByUsername(..., '" + aUsername + "')");
  aListener.onError(null, aUsername, null);
};

/**
 * void getAlbums(in flockIListener aListener, in AString aUsername);
 * @see flockIMediaWebService#findByUsername
 */
DiggService.prototype.getAlbums =
function DiggService_getAlbums(aListener, aUsername) {
  this._logger.debug(".getAlbums(..., '" + aUsername + "')");
  aListener.onError(null, aUsername, null);
};

// DEFAULT: flockIMediaChannel getChannel(in AString aChannelId);

// DEFAULT: nsISimpleEnumerator enumerateChannels();

// DEFAULT: flockIMediaItemFormatter getMediaItemFormatter();

/**
 * void search(in flockIListener aListener, in AString aQuery,
               in long aCount, in long aPage, in AString aRequestId);
 * @see flockIMediaWebService#search
 */
DiggService.prototype.search =
function DiggService_search(aListener, aQuery, aCount, aPage, aRequestId) {
  this._logger.debug(".search(..., '" + aQuery + "', '" + aCount + "', '"
                                      + aPage + "', '" + aRequestId +  "')");
  var query = new queryHelper(aQuery);

  // If the search string only contains whitespace, then return empty result
  var querySearch = query.getKey("search");
  if (querySearch && (querySearch.replace(/\s/g, "") == "")) {
    aListener.onSuccess(createEnum([]), aRequestId);
    return;
  }

  var inst = this;
  var searchListener = {
    onSuccess: function sl_onSuccess(aSubject, aTopic) {
      inst._logger.debug("sl_onResult()");
      var resultsArray = inst._handleUserResults(aSubject.wrappedJSObject);
      aListener.onSuccess(createEnum(resultsArray), aRequestId);
    },
    onError: function sl_onError(aSubject, aTopic, aError) {
      inst._logger.debug("sl_onError()");
      aListener.onError(null, aRequestId, aError);
    }
  };

  // Digg uses a 1-based page index
  var page = aPage;
  if (aPage == 0) {
    page = 1;
  }
  var params = {
    count: aCount,
    offset: (page - 1) * aCount
  };

  var method;
  if (query.hasKey("special")) {
    var specialArray = query.getKey("special").split("-");
    var container = specialArray[0];
    var timeFrame = specialArray[1];
    if (container == "all" ||
        container == "news" ||
        container == "videos" ||
        container == "images")
    {
      method = "stories/popular";
    } else {
      method = "stories/container/" + container + "/popular";
    }

    if (container == "news" || container == "videos" || container == "images") {
      params.media = container;
    }

    if (timeFrame) {
      // Hash table that converts timeFrame into number of seconds
      var timeFrameHash = {
        "24h": 86400,
        "7d":   604800,
        "30d":  2592000,
        "365d": 31536000
      };

      var maxSubmitDate = Math.round(new Date().getTime() / 1000);
      var minSubmitDate = maxSubmitDate - timeFrameHash[timeFrame];
      params.max_submit_date = maxSubmitDate;
      params.min_submit_date = minSubmitDate;
      params.sort = "digg_count-desc";
    }
  } else {
    method = "user/" + query.user + "/dugg";
  }

  gApi.call(method, { wrappedJSObject: params }, searchListener);
};

DiggService.prototype._handleUserResults =
function DiggService__handleUserResults(aJson) {
  var stories = [];
  if (aJson.stories) {
    var inst = this;
    for (var i = 0; (i < aJson.stories.length); i++) {
      var newMediaItem = CC["@flock.com/photo;1"]
                         .createInstance(CI.flockIMediaItem);
      newMediaItem.init(inst.shortName, flockMediaItemFormatter);
      newMediaItem.webPageUrl = addTrackingCode(aJson.stories[i].href);
      var thumb;
      if (aJson.stories[i].thumbnail) {
       thumb = aJson.stories[i].thumbnail.src;
      } else {
       thumb = "chrome://flock/skin/services/digg/diggNoImage.png";
      }
      newMediaItem.thumbnail = thumb;
      newMediaItem.midSizePhoto = thumb;
      newMediaItem.largeSizePhoto = thumb;
      var name = aJson.stories[i].user.name;
      var fullname = aJson.stories[i].user.fullname;
      newMediaItem.username = (fullname) ? fullname : name;
      newMediaItem.userid = name;
      var icon;
      if (aJson.stories[i].user.icon == DIGG_NO_AVATAR_ICON) {
       icon = FLOCK_NO_AVATAR_ICON;
      } else {
       icon = aJson.stories[i].user.icon;
      }
      newMediaItem.icon = icon;
      newMediaItem.is_public = true;
      newMediaItem.is_video = false;

      switch (aJson.stories[i].media) {
        case "news":
          newMediaItem.mediaType = CI.flockIMediaItem.MEDIA_TYPE_NEWS;
          break;
        case "images":
          newMediaItem.mediaType = CI.flockIMediaItem.MEDIA_TYPE_IMAGE;
          break;
        case "videos":
          newMediaItem.mediaType = CI.flockIMediaItem.MEDIA_TYPE_VIDEO;
          break;
        case "podcasts":
          newMediaItem.mediaType = CI.flockIMediaItem.MEDIA_TYPE_PODCAST;
          break;
        default:
          this._logger.debug("No media type found for media item for - "
                             + aJson.stories[i].media + "\n");
          break;
      }

      newMediaItem.title = aJson.stories[i].title;
      newMediaItem.id = aJson.stories[i].id;
      newMediaItem.uploadDate = aJson.stories[i].submit_date * 1000;
      newMediaItem.rating_count = aJson.stories[i].diggs;
      newMediaItem.comment_count = aJson.stories[i].comments;
      stories.push(newMediaItem);
    }
  }
  return stories;
};

/**
 * void supportsSearch(in AString aQuery);
 * @see flockIMediaWebService#supportsSearch
 */
DiggService.prototype.supportsSearch =
function DiggService_supportsSearch(aQuery) {
  return false;
};

/**************************************************************************
 * DiggService: flockIPollingService Implementation
 **************************************************************************/

/**
 * void refresh(in AString aUrn, in flockIPollingListener aListener);
 * @see flockIPollingService#refresh
 */
DiggService.prototype.refresh =
function DiggService_refresh(aUrn, aPollingListener) {
  this._logger.info(".refresh('" + aUrn + "', ...)");
  var acct = this.getAccount(aUrn)
                 .QueryInterface(CI.flockISocialWebServiceAccount);

  if (acct) {
    this._logger.debug("refreshing an Account");
    if (acct.isAuthenticated()) {
      this._refreshAccount(aUrn, aPollingListener);
    } else {
      // If the user is not logged in, return a success without
      // refreshing anything
      aPollingListener.onResult();
    }
  } else {
    this._logger.error("can't refresh " + aUrn + " (unsupported type)");
    aPollingListener.onError(null);
  }
};

/**************************************************************************
 * DiggService: flockIReaderExtension Implementation
 **************************************************************************/
DiggService.prototype.styleSheetUrl = DIGG_READER_CSS_URL;

DiggService.prototype.createFooterElement =
function DiggService_createFooterElement(aDocument, aURI, aListener) {
  var inst = this;
  var bundle = this.getStringBundle();
  var checkDiggInfo_Listener = {
    onSuccess: function getDiggInfo_onSuccess(aResult, aTopic) {
      aResult.QueryInterface(CI.nsIPropertyBag);
      // Container for digg output
      var diggContainer = aDocument.createElement("div");
      diggContainer.setAttribute("id", "digg-container");
      if (aResult.getProperty("diggs") &&
          aResult.getProperty("diggs") == "0")
      {
        // No diggs; but contains a "digg it" URL
        var diggItUrl = aResult.getProperty("diggUrl");
        var diggItTxt = bundle.GetStringFromName("flock.feed.streamReader.digg.diggIt");
        var diggItLink = aDocument.createElement("a");
        diggItLink.setAttribute("href", diggItUrl);
        diggItLink.innerHTML = diggItTxt;

        var diggSpanDiggIt = aDocument.createElement("span");
        diggSpanDiggIt.setAttribute("id", "digg-diggit");
        diggSpanDiggIt.appendChild(diggItLink);
        diggContainer.appendChild(diggSpanDiggIt);
      } else {
        var diggs = aResult.getProperty("diggs");
        var link = aResult.getProperty("link");
        if (diggs && diggs != "0") {
          // Number of diggs
          var diggBundle;
          var diggTxt
          if (diggs == "1") {
            diggBundle = "flock.feed.streamReader.digg.digg";
            diggTxt = bundle.GetStringFromName(diggBundle);
          } else {
            diggBundle = "flock.feed.streamReader.digg.diggs";
            diggTxt = bundle.formatStringFromName(diggBundle, [diggs], 1);
          }
          var diggCount = aDocument.createElement("a");
          diggCount.setAttribute("href", link);
          diggCount.innerHTML = diggTxt;

          // Number of comments
          var commentCount = aDocument.createElement("a");
          var commentBundle = "flock.feed.streamReader.digg.comments";
          var parameters = [aResult.getProperty("comments")];
          var commentTxt = bundle.formatStringFromName(commentBundle,
                                                       parameters,
                                                       1);
          commentCount.setAttribute("href", link);
          commentCount.innerHTML = commentTxt;

          // Assemble and append output
          var diggSpanDiggs = aDocument.createElement("span");
          diggSpanDiggs.setAttribute("id", "digg-diggs");
          diggSpanDiggs.appendChild(diggCount);

          var diggSpanComments = aDocument.createElement("span");
          diggSpanComments.setAttribute("id", "digg-comments");
          diggSpanComments.appendChild(commentCount);

          diggContainer.appendChild(diggSpanDiggs);
          diggContainer.appendChild(diggSpanComments);
        }
      }
      aListener.onSuccess(diggContainer, null);
    },
    onError: function getDiggInfo_onError(aResult, aTopic, aError) {
      inst._logger.error("checkDiggInfo_Listener_onError");
      aListener.onError(aResult, aTopic, aError);
    }
  };

  // Check to see if the feed item is from digg.com
  if (aURI.spec.indexOf(DIGG_URL) != -1) {
    // Digg URLs come in the following form: http://digg.com/{cleanTitle}
    // Extract URL for use in a 'Clean Title' API call.
    var cleanTitle = aURI.spec.substring(aURI.spec.lastIndexOf("/") + 1);
    this.getDiggInfoByCleanTitle(cleanTitle, checkDiggInfo_Listener);
  } else {
    this.loadDiggInfo(aURI.spec, checkDiggInfo_Listener, false);
  }
};


/**************************************************************************
 * DiggService: private members
 **************************************************************************/

DiggService.prototype._refreshAccount =
function DiggService__refreshAccount(aUrn, aPollingListener) {
  this._logger.info("._refreshAccount('" + aUrn + "', ...)");
  var inst = this;
  var account = this.getAccount(aUrn);

  var pplRefreshListener = CC["@flock.com/people-refresh-listener;1"]
                           .createInstance(CI.flockIPeopleRefreshListener);
  pplRefreshListener.init(3, aUrn, aPollingListener, SHORT_INTERVAL);

  // This listener handles getting the user's friends' information.
  var getFriends_Listener = {
    onSuccess: function getFriends_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getFriends_onSuccess(..., '" + aTopic + "')");

      // Now that we have all the results we need, update the RDF.
      var friends = aSubject.wrappedJSObject;
      inst._logger.debug(" - got " + friends.length + " friends");
      // Add friends to list.
      inst._addFriends(account, friends);

      // REMOVE locally people removed on the server
      var localFriends = account.enumerateFriends();
      while (localFriends.hasMoreElements()) {
        var identity = localFriends.getNext();
        if (!friends[identity.accountId]) {
          inst._logger.debug("Friend " + identity.accountId
                             + " has been deleted on the server");
          inst._ppUtils.removePerson(account, identity);
        }
      }
      pplRefreshListener.onSuccess(null, "getFriends_Listener");
    },
    onError: function getFriends_onError(aSubject, aTopic, aError) {
      inst._logger.error("getFriends_onError(..., '" + aTopic + "', ...)");
      pplRefreshListener.onError(aSubject, "getFriends_Listener", aError);
    }
  };

  var getUser_Listener = {
    onSuccess: function getUser_onSuccess(aSubject, aTopic) {
      inst._logger.debug("getUser_onSuccess(..., '" + aTopic + "')");
      aSubject.QueryInterface(CI.nsIPropertyBag);
      account.setParam("name", aSubject.getProperty("name"));
      account.setParam("screenName", aSubject.getProperty("screenName"));
      var avatar = "";
      if (aSubject.getProperty("avatar") != DIGG_NO_AVATAR_ICON) {
        avatar = aSubject.getProperty("avatar");
      }
      account.setParam("avatar", avatar);
      account.setParam("URL", aSubject.getProperty("URL"));
      pplRefreshListener.onSuccess(null, "getUser_Listener");
    },
    onError: function userShowApiListener_onError(aSubject, aTopic, aError) {
      inst._logger.error("getUser_onError(..., '" + aTopic + "', ...)");
      pplRefreshListener.onError(aSubject, "getUser_Listener", aError);
    }
  };

  var uid = account.getParam("accountId");
  this._getMyShouts(pplRefreshListener);
  gApi.getUser(uid, getUser_Listener);
  gApi.getFriends(uid, getFriends_Listener);
};

DiggService.prototype._getMyShouts =
function DiggService__getMyShouts(aPplRefreshListener) {
  this._logger.debug("._getMyShouts()");
  var account = this.getAuthenticatedAccount();
  var inst = this;

  var hr = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
           .createInstance(CI.nsIXMLHttpRequest);
  hr.onreadystatechange = function getMyShouts_onreadystatechange(aEvent) {
    if (hr.readyState == 4) {
      var results = FlockSvcUtils.newResults();
      if (inst._WebDetective.detectNoDOM("digg", "shouts", "",
                                         hr.responseText, results))
      {
        var shoutCount = results.getPropertyAsAString("shoutCount");
        inst._logger.debug("new shouts - " + shoutCount);
        if (account.getParam("accountMessages") < shoutCount) {
          inst._ppUtils.lightPeopleIcon();
        }
        account.setParam("accountMessages", shoutCount);
      } else {
        account.setParam("accountMessages", 0);
      }
      aPplRefreshListener.onSuccess(null, "getMyShouts_onreadystatechange");
    }
  };
  var url = this._WebDetective
                .getString(CLASS_SHORT_NAME, "profileURL", null)
                .replace("%accountid%", account.getParam("accountId"));
  hr.open("GET", url);
  hr.send(null);
};

DiggService.prototype._addFriends =
function DiggService__addFriends(aAccount, aFriends) {
  // ADD or update existing people.
  var friendList = [];
  var lastFriendRefresh = aAccount.getCustomParam("lastFriendRefresh");
  var oneHrAgo = Math.round(Date.now()/1000) - FRIEND_INTERVAL_FETCH;
  var refreshTimespan = Math.abs(oneHrAgo - lastFriendRefresh);
  if (refreshTimespan > FRIEND_INTERVAL_FETCH) {
    // Last refresh was greater than max, do a non-batch refresh.
    this._nonBatchAddFriends(aAccount, aFriends);
  } else {
    this._addBatchFriend(aAccount, aFriends);
  }
  aAccount.setCustomParam("lastFriendRefresh", Math.round(Date.now()/1000));
};

DiggService.prototype._nonBatchAddFriends =
function DiggService__nonBatchAddFriends(aAccount, aFriends) {
  for (var f in aFriends) {
    this._logger.debug("about to add - " + f);
    var person = this._getPerson(aAccount, aFriends[f]);
    this._updateFriend(aAccount, person);
  }
};

DiggService.prototype._updateFriend =
function DiggService__updateFriend(aAccount, aFriend) {
  // Update friends Activity.
  var inst = this;
  var getFriendActivity_Listener = {
    onSuccess: function getFriendActivity_onSuccess(aSubject, aTopic) {
      inst._ppUtils.addPerson(aAccount, aSubject);
    },
    onError: function getFriendActivity_onError(aSubject, aTopic, aError) {
      inst._logger.error("getFriendActivity_onError('" + aTopic + "')");
      inst._ppUtils.addPerson(aAccount, aSubject);
    }
  };
  gApi.getFriendActivity(aAccount, aFriend, getFriendActivity_Listener);
};

DiggService.prototype._addBatchFriend =
function DiggService__addBatchFriend(aAccount, aFriends) {
  var inst = this;
  var getFriendActivity_Listener = {
    onSuccess: function getFriendActivity_onSuccess(aSubject, aTopic) {
      inst._logger.info("_addBatchFriend on Success(...,'" + aTopic + "')");
      inst._ppUtils.addPerson(aAccount, aSubject);
    },
    onError: function getFriendActivity_onError(aSubject, aTopic, aError) {
      inst._logger.error("_addBatchFriend getFriendActivity_onError('"
                         + aTopic + "')");
      inst._ppUtils.addPerson(aAccount, aSubject.wrappedJSObject);
    }
  };
  var friendsList = CC["@mozilla.org/hash-property-bag;1"]
                    .createInstance(CI.nsIWritablePropertyBag);
  for (i in aFriends) {
    var friend = this._getPerson(aAccount, aFriends[i]);
    // Does Flock know about this friend?
    if (this._isFriendAlready(aAccount, aFriends[i])) {
      friendsList.setProperty(friend.accountId, friend);
    } else {
      // This is a new friend
      this._updateFriend(aAccount, friend);
    }
  }

  gApi.getBatchFriendActivity(aAccount, friendsList, getFriendActivity_Listener);
};

DiggService.prototype._getIdentityUrn =
function DiggService__getIdentityUrn(aCoopAccount, aPersonId) {
  var coopService = aCoopAccount.service;
  var result = "urn:flock:identity:"
             + coopService.name.toLowerCase() + ":"
             + aCoopAccount.accountId + ":"
             + aPersonId;

  return result;
}

DiggService.prototype._isFriendAlready =
function DiggService__isFriendAlready(aAccount, aPerson) {
  // Does this friend already exist in the RDF?
  var personName = aPerson.getProperty("name");
  var coopAccount = this._coop.get(aAccount.urn);
  var identityUrn = this._getIdentityUrn(coopAccount, personName);
  return this._coop.Identity.exists(identityUrn);
};

DiggService.prototype._getPerson =
function DiggService__getPerson(aAccount, aPerson) {
  var inst = this;
  var person = null;
  var personName = aPerson.getProperty("name");
  var coopAccount = this._coop.get(aAccount.urn);
  var identityUrn = this._getIdentityUrn(coopAccount, personName);
  if (this._coop.Identity.exists(identityUrn)) {
    var coopPerson = this._coop.get(identityUrn);
    person = CC["@flock.com/person;1"]
               .createInstance(CI.flockIPerson);
    person.accountId = coopPerson.accountId;
    person.lastUpdate = coopPerson.lastUpdate;
    person.lastUpdateType = coopPerson.lastUpdateType;
    person.name = coopPerson.name;
    person.avatar = coopPerson.avatar;
    person.statusMessage = coopPerson.statusMessage;
    person.profileURL = coopPerson.URL;
    person.screenName = coopPerson.screenName;
  }

  if (!person) {
    person = CC["@flock.com/person;1"]
             .createInstance(CI.flockIPerson);
    aPerson.QueryInterface(CI.nsIPropertyBag);
    person.accountId = personName;
    person.screenName = personName;
    person.lastUpdate = Math.round((Date.now()/1000) - MEDIA_INITIAL_FETCH);
    person.lastUpdateType = "";
    person.name = personName;
    person.avatar = aPerson.getProperty("avatar");

    this._logger.debug("Adding person: '" + aAccount + "' '"
                       + person.accountId + "' '"
                       + "'");
  }
  return person;
};


/*************************************************************************
 * Component: DiggAccount
 *************************************************************************/
function DiggAccount() {
  FlockSvcUtils.getLogger(this);
  this._logger.init("DiggAccount");

  this._WebDetective = CC["@flock.com/web-detective;1"]
                       .getService(CI.flockIWebDetective);
  FlockSvcUtils.getACUtils(this);
  FlockSvcUtils.getCoop(this);

  var wsa = FlockSvcUtils.flockIWebServiceAccount;
  wsa.addDefaultMethod(this, "getAllCustomParams");
  wsa.addDefaultMethod(this, "getCustomParam");
  wsa.addDefaultMethod(this, "getParam");
  wsa.addDefaultMethod(this, "getService");
  wsa.addDefaultMethod(this, "isAuthenticated");
  wsa.addDefaultMethod(this, "logout");
  wsa.addDefaultMethod(this, "setCustomParam");
  wsa.addDefaultMethod(this, "setParam");

  var sa = FlockSvcUtils.flockISocialWebServiceAccount;
  sa.addDefaultMethod(this, "enumerateFriends");
  sa.addDefaultMethod(this, "getFriendCount");
}

// Use genericComponent() for the goodness it provides (QI, nsIClassInfo, etc),
// but do NOT add this component to the list of constructors passed to
// FlockXPCOMUtils.genericModule().
DiggAccount.prototype = new FlockXPCOMUtils.genericComponent(
  CLASS_NAME + " Account",
  "",
  "",
  DiggAccount,
  0,
  [
    CI.flockIWebServiceAccount,
    CI.flockIDiggAccount,
    CI.flockISocialWebServiceAccount,
    CI.flockIMediaWebServiceAccount
  ]
);


/*************************************************************************
 * DiggAccount: flockIWebServiceAccount Implementation
 *************************************************************************/

// readonly attribute AString urn;
DiggAccount.prototype.urn = "";

// DEFAULT: nsIPropertyBag getAllCustomParams();
// DEFAULT: nsIVariant getCustomParam(in AString aParamName);
// DEFAULT: nsIVariant getParam(in AString aParamName);
// DEFAULT: flockILoginWebService getService();
// DEFAULT: boolean isAuthenticated();

/**
 * void keep();
 * @see flockIWebServiceAccount#keep
 */
DiggAccount.prototype.keep =
function DiggAccount_keep() {
  this._logger.debug(".keep()");
  var c_acct = this._coop.get(this.urn);
  c_acct.isTransient = false;
  this._acUtils.makeTempPasswordPermanent("urn:digg:service:"
                                          + c_acct.accountId);
};

/**
 * void login(in flockIListener aListener);
 * @see flockIWebServiceAccount#login
 */
DiggAccount.prototype.login =
function DiggAccount_login(aListener) {
  this._logger.debug(".login()");

  this._logger.debug("{flockIWebServiceAccount}.login()");
  var c_acct = this._coop.get(this.urn);
  if (!c_acct.isAuthenticated) {
    c_acct.isAuthenticated = true;
  }
  // Force refresh on login
  var pollerSvc = CC["@flock.com/poller-service;1"]
                  .getService(CI.flockIPollerService);
  pollerSvc.forceRefresh(this.urn);
  if (aListener) {
    aListener.onSuccess(this, "login");
  }
};

// void activate(in flockIListener aListener); 
DiggAccount.prototype.activate = 
function PDiggAccount_activate(aListener) { 
  this._logger.debug(".activate()"); 
  this.setParam("isAuthenticated", true); 
  this.setParam("nextRefresh", new Date(0)); 
  this.setParam("isPollable", true); 
 
  if (aListener) { 
    aListener.onSuccess(this, "activate"); 
  } 
};

// DEFAULT: void logout(in flockIListener aListener);
// DEFAULT: void setCustomParam(in AString aParamName, in nsIVariant aValue);
// DEFAULT: void setParam(in AString aParamName, in nsIVariant aValue);


/*************************************************************************
 * DiggAccount: flockISocialWebServiceAccount Implementation
 *************************************************************************/

// readonly attribute boolean hasFriendActions;
DiggAccount.prototype.hasFriendActions = true;

// readonly attribute boolean isMeStatusSupported;
DiggAccount.prototype.isMeStatusSupported = false;

// readonly attribute boolean isFriendStatusSupported;
DiggAccount.prototype.isFriendStatusSupported = true;

// readonly attribute boolean isStatusEditable;
DiggAccount.prototype.isStatusEditable = false;

// readonly attribute boolean isPostLinkSupported;
DiggAccount.prototype.isPostLinkSupported = true;

// readonly attribute boolean isMyMediaFavoritesSupported;
DiggAccount.prototype.isMyMediaFavoritesSupported = false;

// DEFAULT: nsISimpleEnumerator enumerateFriends();
/**
 * AString formatFriendActivityForDisplay(in AString aFriendActivityUrn);
 * @see flockISocialAccount#formatFriendActivityForDisplay
 */
DiggAccount.prototype.formatFriendActivityForDisplay =
function DiggAccount_formatFriendActivityForDisplay(aFriendActivityUrn) {
  this._logger.debug("formatFriendActivityForDisplay("
                     + aFriendActivityUrn  + ")");

  var friendActivity = FlockSvcUtils.getCoop(this)
                                    .get(aFriendActivityUrn);
  var bundle = this.getService().getStringBundle();
  var result = "";

  switch (friendActivity.updateType) {
   case "profile":
     result =
       bundle.GetStringFromName("flock.friendFeed.lastUpdateTypePretty.profile");
     break;
 
   case "commented":
     result = bundle.GetStringFromName("flock.friendFeed.lastUpdateTypePretty.commented")
            + " "
            + bundle.GetStringFromName("flock.friendFeed.activity.prefix.commented")
            + " "
            + friendActivity.updateValue;
     break;

   case "dugg":
     result = bundle.GetStringFromName("flock.friendFeed.lastUpdateTypePretty.dugg")
            + " "
            + bundle.GetStringFromName("flock.friendFeed.activity.prefix.dugg")
            + " "
            + friendActivity.updateValue;
     break;

   case "submitted":
     result = bundle.GetStringFromName("flock.friendFeed.lastUpdateTypePretty.submitted")
            + " "
            + bundle.GetStringFromName("flock.friendFeed.activity.prefix.submitted")
            + " "
           + friendActivity.updateValue;
     break;
  }

  return flockXMLDecode(result);
};

/**
 * AString formatStatusForDisplay(in AString aStatusMessage);
 * @see flockISocialWebServiceAccount#formatStatusForDisplay
 */
DiggAccount.prototype.formatStatusForDisplay =
function DiggAccount_formatStatusForDisplay(aStatusMessage) {
  return (aStatusMessage) ? flockXMLDecode(aStatusMessage) : "";
};

/**
 * AString getEditableStatus();
 * @see flockISocialWebServiceAccount#getEditableStatus
 */
DiggAccount.prototype.getEditableStatus =
function DiggAccount_getEditableStatus() {
  return "";
};

/**
 * AString getFriendActions(in AString aFriendURN);
 * @see flockISocialWebServiceAccount#getFriendActions
 */
DiggAccount.prototype.getFriendActions =
function DiggAccount_getFriendActions(aFriendURN) {
  this._logger.info(".getFriendActions('" + aFriendURN + "')");

  var actionNames = ["friendShout",
                     "friendDiggs",
                     "friendComments",
                     "friendSubmissions",
                     "friendFavorites",
                     "friendGallery",
                     "friendShareFlock"];

  var bundle = this.getService().getStringBundle();

  var actions = [];
  var c_friend = this._coop.get(aFriendURN);
  if (c_friend) {
    for each (var i in actionNames) {
      actions.push({
        label: bundle.GetStringFromName("flock.digg.actions." + i),
        class: i,
        spec: this._WebDetective.getString("digg", i, "")
                  .replace("%friendid%", c_friend.accountId)
      });     
    }
  }
  return JSON.toString(actions);
};

// DEFAULT: long getFriendCount();

/**
 * AString getMeNotifications();
 * @see flockISocialWebServiceAccount#getMeNotifications
 */
DiggAccount.prototype.getMeNotifications =
function DiggAccount_getMeNotifications() {
  this._logger.info(".getMeNotifications() " + this.urn);

  var bundle = this.getService().getStringBundle();
  var noties = [];
  var inst = this; 
  function _addNotie(aType, aCount) {
    var stringName = "flock.digg.noties."
                   + aType + "."
                   + ((parseInt(aCount) <= 0) ? "none" : "some");
    noties.push({
      class: aType,
      tooltip: bundle.GetStringFromName(stringName),
      metricsName: aType,
      count: (aCount) ? aCount : 0,
      URL: inst._WebDetective.getString("digg", aType + "_URL", "")
                             .replace("%accountid%", inst.getParam("accountId"))
    });
  }

  var c_params = this._acUtils.getAllAccountCustomParams(this.urn);
  _addNotie("meComments", c_params.getProperty("diggComments"));
  _addNotie("meMessages", this.getParam("accountMessages"));
  return JSON.toString(noties);
};

/**
 * AString getProfileURLForFriend(in AString aFriendURN);
 * @see flockISocialWebServiceAccount#getProfileURLForFriend
 */
DiggAccount.prototype.getProfileURLForFriend =
function DiggAccount_getProfileURLForFriend(aFriendURN) {
  this._logger.info(".getProfileURLForFriend('" + aFriendURN + "')");
  var url = "";
  var c_friend = this._coop.get(aFriendURN);
  if (c_friend) {
    url =  this._WebDetective
               .getString(CLASS_SHORT_NAME, "profileURL", null)
               .replace("%accountid%", c_friend.accountId);
  }
  return url;
};

/**
 * AString getPostLinkAction(in nsITransferable aTransferable);
 * @see flockISocialWebServiceAccount#getPostLinkAction
 */
DiggAccount.prototype.getPostLinkAction =
function DiggAccount_getPostLinkAction(aTransferable) {
  return "method:flockIDiggAccount:diggLink";
};

/**
 * AString diggLink();
 * @see flockIDiggAccount#diggLink
 */
DiggAccount.prototype.diggLink =
function DiggAccount_diggLink() {
  var wm = CC["@mozilla.org/appshell/window-mediator;1"]
           .getService(CI.nsIWindowMediator);
  var win = wm.getMostRecentWindow("navigator:browser");
  if (!win) {
    return;
  }
  var url = win.gBrowser.currentURI.spec;
  var browser = win.getBrowser();
  if (url && (!gDuggCurrentUrl ||
              gDuggCurrentUrl.toLowerCase() != url.toLowerCase()))
  {
    // We never recieve digg information for this information, do it now
    var inst = this;
    var diggLink_Listener = {
      onSuccess: function loc_success(aResult) {
        aResult.QueryInterface(CI.nsIPropertyBag);
        url = aResult.getProperty("diggUrl");
        browser.loadOneTab(url, null, null, null, false, false);
      },
      onError: function loc_error(aError) {
        inst._logger.error("getDiggInfo_onError(..., '" + aError + ", ...)");
      }
    }
    gApi.getDiggInformation(url, diggLink_Listener, false);
  } else {
    browser.loadOneTab(url, null, null, null, false, false);
  }
};

/**
 * AString getSharingAction(in AString aFriendURN,
 *                          in nsITransferable aTransferable);
 * @see flockISocialWebServiceAccount#getSharingAction
 */
DiggAccount.prototype.getSharingAction =
function DiggAccount_getSharingAction(aFriendURN, aTransferable) {
  this._logger.info(".getSharingAction('" + aFriendURN + "')");

  var sharingAction = "";
  var c_friend = this._coop.get(aFriendURN);
  if (c_friend) {
    var flavors = ["text/x-flock-media",
                   "text/x-moz-url",
                   "text/unicode",
                   "text/html"];

    var message = CC[FLOCK_RDNDS_CONTRACTID]
                  .getService(CI.flockIRichDNDService)
                  .getMessageFromTransferable(aTransferable,
                                              flavors.length,
                                              flavors);
    if (!message.body) {
      return sharingAction;
    } else {
      sharingAction = this._WebDetective.getString(CLASS_SHORT_NAME,
                                                   "shareAction_shout",
                                                   "");
    }
  }
  this._logger.info(sharingAction);
  return sharingAction;
};

/**
 * void diggShout(in AString aFriendURN, in nsITransferable aTransferable);
 * @see flockIDiggAccount#diggShout
 */
DiggAccount.prototype.diggShout =
function DiggAccount_diggShout(aFriendURN, aTransferable) {
  this._logger.debug("diggShout");
  var flavors = ["text/x-flock-media",
                 "text/x-moz-url",
                 "text/unicode",
                 "text/html"];

  var message = CC[FLOCK_RDNDS_CONTRACTID]
                .getService(CI.flockIRichDNDService)
                .getMessageFromTransferable(aTransferable,
                                            flavors.length,
                                            flavors);
  if (message.body) {
    this._composeMessage(aFriendURN, message.body, true);
  }
};

DiggAccount.prototype.shareFlock =
function DiggAccount_shareFlock(aFriendURN) {
  this._logger.info(".shareFlock('" + aFriendURN + "')");

  // Get "Share Flock" message text
  var sbs = CC["@mozilla.org/intl/stringbundle;1"]
            .getService(CI.nsIStringBundleService);
  var bundle = sbs.createBundle(DIGG_PROPERTIES);
  var body = "";
  try {
    for (var i = 0; ; i++) {
      body += "\n"
           + bundle.GetStringFromName("flock.digg.friendShareFlock.message"
                                      + i);
    }
  } catch (ex) {
    // Ignore -- we've hit the end of our message lines
  }
  this._composeMessage(aFriendURN, body, false);
}

DiggAccount.prototype._composeMessage =
function DIiggAccount__composeMessage(aFriendURN,
                                      aBody,
                                      addBreadCrumb)
{
  var body = aBody;
  var c_friend = this._coop.get(aFriendURN);
  var url = this._WebDetective.getString(CLASS_SHORT_NAME, "friendShout", "")
                .replace("%friendid%", c_friend.accountId);
  url = url.substring(url.indexOf("http"), url.length);
  var wm = CC["@mozilla.org/appshell/window-mediator;1"]
             .getService(CI.nsIWindowMediator);
  var win = wm.getMostRecentWindow("navigator:browser");
  if (win) {
    var browser = win.getBrowser();
    var newTab = browser.loadOneTab(url, null, null, null, false, false);
    var obs = CC["@mozilla.org/observer-service;1"]
              .getService(CI.nsIObserverService);
    var inst = this;
    var observer = {
      observe: function openSendMessageTabForFill_observer(aContent,
                                                           aTopic,
                                                           aContextUrl)
      {
        var contentWindow = newTab.linkedBrowser.docShell
                                  .QueryInterface(CI.nsIInterfaceRequestor)
                                  .getInterface(CI.nsIDOMWindow);
        function insertContent(aWebDicString, aMessage) {
          var xpathquery = inst._WebDetective.getString(CLASS_SHORT_NAME,
                                                        aWebDicString, "");
          var doc = contentWindow.document;
          var formItems = doc.evaluate(xpathquery, doc, null,
                                       CI.nsIDOMXPathResult.ANY_TYPE, null);
          if (formItems) {
            var formItem = formItems.iterateNext();
            if (formItem.hasAttribute("value")) {
              formItem.setAttribute("value", aMessage);
            } else {
              var textNode = doc.createTextNode(aMessage);
              formItem.appendChild(textNode);
            }
            formItem.focus();
          }
        };
        if (contentWindow == aContent) {
          obs.removeObserver(this, "EndDocumentLoad");
          if (addBreadCrumb) {
            // Add breadcrumb to message body
            var breadcrumb = CC[FLOCK_RDNDS_CONTRACTID]
                             .getService(CI.flockIRichDNDService)
                             .getBreadcrumb("plain");
            if (breadcrumb) {
              body += breadcrumb;
            }
          }
          insertContent("diggmessage_sendXPath", body);
        }
      }
    };
    obs.addObserver(observer, "EndDocumentLoad", false);
  }
};

/**
 * void markAllMeNotificationsSeen(in AString aType);
 * @see flockISocialWebServiceAccount#markAllMeNotificationsSeen
 */
DiggAccount.prototype.markAllMeNotificationsSeen =
function DiggAccount_markAllMeNotificationsSeen(aType) {
  this._logger.debug(".markAllMeNotificationsSeen('" + aType + "')");
  switch (aType) {
    case "meMessages":
      this.setParam("accountMessages", 0);
      break;  
    case "meComments":
      this.setCustomParam("diggComments", 0);
      break;  
    default:
      break;
  }
};

/**
 * void setStatus(in AString aStatusMessage, in flockIListener aListener);
 * @see flockISocialWebServiceAccount#setStatus
 */
DiggAccount.prototype.setStatus =
function DiggAccount_setStatus(aStatusMessage, aListener) {
  throw CR.NS_ERROR_NOT_IMPLEMENTED;
};

DiggAccount.prototype.getFormattedFriendUpdateType =
function DiggAccount_getFormattedFriendUpdateType(aLastUpdateType, aStatus) {
  if (!aLastUpdateType) {
    return "";
  }
  var bundleSubString = "flock.friendFeed.lastUpdateTypePretty."
                      + aLastUpdateType;
  this._logger.info("getFormattedFriendUpdateType " + bundleSubString);
  var bundle = this.getService().getStringBundle();
  return bundle.GetStringFromName(bundleSubString);
};


/*************************************************************************
 * XPCOM Support - Module Construction
 *************************************************************************/

// Create array of components.
var componentsArray = [DiggAPI, DiggService];

// Generate a module for XPCOM to find.
var NSGetModule = FlockXPCOMUtils.generateNSGetModule(MODULE_NAME,
                                                      componentsArray);
